#pragma once

// From Vista headers.
#if (NTDDI_VERSION < NTDDI_VISTA)
enum _LI_METRIC
{
   LIM_SMALL, // corresponds to SM_CXSMICON/SM_CYSMICON
   LIM_LARGE, // corresponds to SM_CXICON/SM_CYICON
};
#endif

namespace LeoHelpers
{
	BOOL GetDlgItemRect(HWND hwndDlg, int nIDDlgItem, RECT *pRect);

	BOOL OffsetDlgItem(HWND hwndDlg, int nIDDlgItem, LONG dx, LONG dy, LONG dw, LONG dh, BOOL bRepaint);

	BOOL EnableDlgItem(HWND hwndDlg, int nIDDlgItem, BOOL bEnable);
	BOOL ShowDlgItem(HWND hwndDlg, int nIDDlgItem, BOOL bShow);

	BOOL GetDlgItemText(HWND hwndDlg, int nIDDlgItem, std::wstring *pStrResult);
	BOOL GetListBoxItemText(HWND hwndListBox, LRESULT lrItemIndex, std::wstring *pStrResult);
	BOOL SetupComboItem(HWND hWndParent, int iDlgItem, bool bClear, DWORD dwValue, DWORD dwSelectValue, const wchar_t *szTitle);
	BOOL SetComboItem(HWND hWndParent, int iDlgItem, DWORD dwValue);
	BOOL GetComboItem(HWND hWndParent, int iDlgItem, DWORD *pdwValue);
	BOOL GetComboItemIntCast(HWND hWndParent, int iDlgItem, int *piValue);
	BOOL SetupCheckBoxItem(HWND hWndParent, int iDlgItem, bool bCheck);
	BOOL GetCheckBoxItem(HWND hWndParent, int iDlgItem, bool *pbValue);
	BOOL SetupSpinItem(HWND hWndParent, int iDlgItem, int iMax, int iMin, int iValue);
	BOOL SetSpinItem(HWND hWndParent, int iDlgItem, int iValue);
	BOOL GetSpinItem(HWND hWndParent, int iDlgItem, int *piValue);
	BOOL SetupEditItem(HWND hWndParent, int iDlgItem, const wchar_t *szText);
	BOOL GetEditItem(HWND hWndParent, int iDlgItem, std::wstring *pStrResult);

	bool FilenameToEdit(HWND hwndDlg, int iDlgItem, const wchar_t *szTitle, const wchar_t *szFilter, bool bAddToRecent, DWORD *pdwCommDlgExtendedError);

//	BOOL SubclassDlgItemToUserData(HWND hwndDlg, int nIDDlgItem, WNDPROC wndProc);

//	BOOL GetDlgItemLong(HWND hwndDlg, int nIDDlgItem, int nIndex, LONG *plResult);
//	BOOL SetDlgItemLong(HWND hwndDlg, int nIDDlgItem, int nIndex, LONG lValue);
	BOOL GetDlgItemLongPtr(HWND hwndDlg, int nIDDlgItem, int nIndex, LONG_PTR *plResult);
	BOOL SetDlgItemLongPtr(HWND hwndDlg, int nIDDlgItem, int nIndex, LONG_PTR lValue);
//	BOOL ModifyDlgItemStyle(HWND hwndDlg, int nIDDlgItem, bool bStyleEx, LONG_PTR lAdd, LONG_PTR lRemove);

	LRESULT SendMessageDlgItem(HWND hwndDlg, int nIDDlgItem, UINT Msg, WPARAM wParam, LPARAM lParam);

	BOOL InvalidateDlgItem(HWND hwndDlg, int nIDDlgItem);

	BOOL EnableDlgControl(HWND hwndControl, BOOL bEnable);

	HRESULT LoadIconMetricOrDefault(HINSTANCE hinst, PCWSTR pszName, int lims, HICON *phico, bool *pbDestroyIcon);

	// ******************************************************************************************************
	// This ThemeHelper class is NOT THREAD SAFE
	// If you have multiple UI threads sharing the same object instance (WHY?) then use some kind of locking.
	// ******************************************************************************************************
	class ThemeHelper
	{
	public:
		ThemeHelper();
		~ThemeHelper(); // non-virtual

	private:

		ThemeHelper(const ThemeHelper &rhs); // disallow
		ThemeHelper &operator=(const ThemeHelper &rhs); // disallow

	public:
		bool CacheBackgroundTabThemeDialog(HWND hWnd);

		// Returns TRUE iff it was enabled.
		bool EnableThemeDialogTexture(HWND hWnd, DWORD dwFlags);

		// You have to call EnableThemeDialogTexture and have Windows erase the background in order to get
		// "transparent" controls if version 6 common-controls are enabled. On XP the tab dialog texture
		// is a limited height and runs out if the window is sized too tall. This works around that problem
		// by tiling the texture, flipping it over on every alternate line so that it remains smooth.
		// Some child controls may paint their own backgrounds based on what they think the parent window's
		// background should be, so some "transparent" controls may need subclassing to paint properly below
		// the point where the background texture runs out on XP.
		// On Vista the texture seems to tile correctly regardless of the dialog size without needing help.
		void EraseBackgroundTabThemeDialog(HWND hWnd, HDC hDC, const RECT &rcClient, const RECT &rcBackgroundWnd);

		HBRUSH GetTempBrush(HWND hWnd, HDC hDC, const RECT &rcClient, const RECT &rcBackgroundWnd);

//		BOOL IsThemeBackgroundPartiallyTransparentButton(HWND hWndControl, int iPartId, int iStateId);

		HRESULT DrawThemeParentBackground(HWND hWnd, HDC hDC, const RECT *prcDrawChildCoords);

	protected:
		SIZE m_sizePart;
		HBITMAP m_hbmpDIB;

		HDC m_hdcMemTemp;
		HBRUSH m_hBrushTemp;
	};

	class CommonDialogData
	{
	private:
		bool                    m_bShuttingDown;
		LeoHelpers::ThemeHelper m_themeHelper;
		void *                  m_pvUserData;
		std::vector< HWND >     m_tabBackgroundDlgs;
		std::vector< HWND >     m_eraseBackgroundCallers;

	public:
		CommonDialogData(void *pvUserData);
		virtual ~CommonDialogData();

	private:
		CommonDialogData(const CommonDialogData &rhs); // disallow
		CommonDialogData &operator=(const CommonDialogData &rhs); // disallow

	public:
		LeoHelpers::ThemeHelper &GetThemeHelper() { return m_themeHelper; }
		void *GetUserData() const { return m_pvUserData; }

		void AddTabBackgroundDlg(HWND hwndDlg);

		// If you have any tabs your main dialog proc should call here as part of its WM_THEMECHANGED
		// processing. You should also make one call during initialisation. Before any calls,
		// use AddTabBackgroundDlg to specify the child dialogs that need tab-theme backgrounds.
		void OnThemeChanged();

		// If you have a tab that contains a child dialog, the child dialog's WM_ERASEBKGND should call this
		// passing its own window handle and the HDC passed to WM_ERASEBKGND's wParam argument, and return
		// the result.
		BOOL TabChildDlgEraseBackground(HWND hwndDlg, HDC hDC);

		// If you have a tab that contains a child dialog, the child dialog's WM_CTLCOLORBTN and WM_CTLCOLORSTATIC
		// should call this passing its own window handle and the HDC passed to the messages' wParam arguments,
		// and return the result (cast to INT_PTR).
		HBRUSH TabChildDlgCtlColor(HWND hwndDlg, HDC hDC);

		// You can use these to install the subclasses, or you can install or call them yourself (e.g. if you want to
		// add additional subclass behaviour on top of that provided by these library functions).

		// These methods install the subclass functions in the default way.

		// Subclassing with the DoubleBufferPaintProc can remove flicker on resize and ensure tab-background textures are painted & aligned correctly.
		void SubclassDoubleBufferPaintProc(HWND hwndCtl);

		// Subclass with the BackgroundAlignPaintProc can ensure tab-background textures are aligned correctly.
		void SubclassBackgroundAlignPaintProc(HWND hwndCtl);

		// Subclassing with the NoEraseListBoxPaintProc can remove flicker when resizing listbox controls, in their empty areas.
		void SubclassNoEraseListBoxPaintProc(HWND hwndCtl);

	protected:
		static LRESULT CALLBACK staticNoEraseListBoxPaintProc(HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam, UINT_PTR uIdSubclass, DWORD_PTR dwRefData);
		static LRESULT CALLBACK staticBackgroundAlignPaintProc(HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam, UINT_PTR uIdSubclass, DWORD_PTR dwRefData);
		static LRESULT CALLBACK staticDoubleBufferPaintProc(HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam, UINT_PTR uIdSubclass, DWORD_PTR dwRefData);

		LRESULT memberNoEraseListBoxPaintProc(HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam);
		LRESULT memberBackgroundAlignPaintProc(HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam);
		LRESULT memberDoubleBufferPaintProc(HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam);

		LRESULT onPaintDoubleBuffer(HWND hWnd);
		void doubleBufferPaintContents(HDC hDC, HWND hWnd, HWND hWndParent, const RECT &rcParent, const PAINTSTRUCT &paint);
	};
};
